function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end


-- dofile "/Projects/Hearts/ALL/Scripts/copytable.lua"
-- require "copytable"

-----------------------------------------------------------------------------------------------------------
-- Variables
-----------------------------------------------------------------------------------------------------------

ace = 14
king = 13
queen = 12
jack = 11
ten = 10
nine = 9
eight = 8
seven = 7
six = 6
five = 5
four = 4
three = 3
two = 2
ace_low = 1

hearts = 1
spades = 2
diamonds = 3
clubs = 4

suit_strs  = { "Hearts","Spades","Diamonds","Clubs" }
suit_chars = { "h","s","d","c" }
rank_strs  = { "Ace (Low)","Two","Three","Four","Five","Six","Seven","Eight","Nine","Ten","Jack","Queen","King","Ace" }
rank_chars = { "a","2","3","4","5","6","7","8","9","T","J","Q","K","A" }

-----------------------------------------------------------------------------------------------------------
-- RS
-----------------------------------------------------------------------------------------------------------

function RS(rs_r,rs_s)

	if not rs_r or not rs_s then print("RS(): no rank or suit (",rs_r,rs_s,")") return end

	rs_c = { rank = rs_r, suit = rs_s }
	return rs_c
	
end


-----------------------------------------------------------------------------------------------------------
-- CardGroup Class
-----------------------------------------------------------------------------------------------------------

--Card =
--{
--	suit = nil,
--	rank = nil
--}



CardGroup =
{
	name      = "Unset",
	name_num  = 0,
	num_cards = 0,
	cards     =  {}
}

function CardGroup.Create(name,name_num)

	if (show_calls) then print("CardGroup.Create:",name,name_num) end
	
	local self = deepcopy(CardGroup)
	
	self.num_cards = 0
	self.name      = name
	self.name_num  = name_num
	
	if (name_num) then self.name = self.name .. tostring(name_num) end

	if (show_calls) then print("CardGroup.Create: DONE") end
	
	return self

end

function CardGroup:AddCardRS(r,s)

	if (not r or r < 1  or r > ace) then print("Error - CardGroup:AddCardRS() - invalid rank:",r) return end
	if (not s or s < 1  or s > 4)   then print("Error - CardGroup:AddCardRS() - invalid suit:",s) return end

	if (show_calls) then print("["..self.name.."] CardGroup.AddCardRS: r=",r,"s=",s) end

	self.num_cards = self.num_cards + 1
	self.cards[self.num_cards] = { rank = r, suit = s }

end

function CardGroup:AddCardRSNoCheck(r,s)

	if (show_calls) then print("["..self.name.."] CardGroup.AddCardRSNoCheck: r=",r,"s=",s) end

	self.num_cards = self.num_cards + 1
	self.cards[self.num_cards] = { rank = r, suit = s }

end

function CardGroup:AddCard(c)

	if (not c) then print("Error - CardGroup:AddCard() - no card") return end
	
	if (show_calls) then print("CardGroup.AddCard: c=",c.rank,c.suit) end
	
	self:AddCardRS(c.rank,c.suit)

end

function CardGroup:HasCardRS(r,s)

	if (not r or not s) then print("Error - CardGroup:HasCardRS() - no suit or rank") return end

	for i = 1,self.num_cards do
		c = self.cards[i]
		if (c.rank == r and c.suit == s) then return true end
	end

	return false
end

function CardGroup:HasCard(c)

	if (not c) then print("Error - CardGroup:HasCard() - no card") return end

	return self:HasCardRS(c.rank,c.suit)

end

function CardGroup:GetCardRS(r,s)

	if (not r or not s) then print("Error - CardGroup:GetCardRS() - no suit or rank") return end

	for i = 1,self.num_cards do
		c = self.cards[i]
		if (c.rank == r and c.suit == s) then return c end
	end

	return false
end

function CardGroup:GetCard(c)

	if (not c) then print("Error - CardGroup:GetCard() - no card") return end

	return self:GetCardRS(c.rank,c.suit)

end

function CardGroup:RemoveCardRS(r,s)

	if (not r or r < 1  or r > ace) then print("Error - CardGroup:RemoveCardRS() - invalid rank:",r) return end
	if (not s or s < 1  or s > 4)   then print("Error - CardGroup:RemoveCardRS() - invalid suit:",s) return end

	if (show_calls) then print("["..self.name.."] CardGroup.RemoveCardRS: r=",r,"s=",s) end

	cnum = 0
	for i=1,self.num_cards do
		c = self.cards[i]
		if (c.rank == r and c.suit == s) then cnum = i end
	end

	if (cnum == 0) then print("Error - CardGroup:RemoveCardRS() - card: "..rank_chars[r]..suit_chars[s].." not found") return end

	if (cnum == self.num_cards) then
		self.cards[self.num_cards] = nil
		self.num_cards = self.num_cards - 1
		return
	end
	
	for i=cnum+1,self.num_cards do
		self.cards[i-1].rank = self.cards[i].rank
		self.cards[i-1].suit = self.cards[i].suit
	end

	self.cards[self.num_cards] = nil
	self.num_cards = self.num_cards - 1
	
end

function CardGroup:RemoveCard(c)

	if not c then print("ERROR: CardGroup:RemoveCard - no card") return end
	
	self:RemoveCardRS(c.rank,c.suit)

end

function CardGroup:RemoveAllCards()

	for i=1,self.num_cards do
		self.cards[i] = nil
	end
	
	self.num_cards = 0

end

function CardGroup:CopyFromOtherCardGroup(g)

	if (not self) then print("Error> CardGroup:CopyFromOtherCardGroup() - no group") return end
	
	self:RemoveAllCards()
	
	if (not g) then print("Error> CardGroup:CopyFromOtherCardGroup() - no group to copy") return end
	
	for i=1,g.num_cards do
		self:AddCard(g.cards[i])
	end

end

function CardGroup:SortCardsByRank(dir)

	if (not self) then print("Error> CardGroup:SortCardsByRank() - no group") return end
	
	if (dir and dir < 0) then
		table.sort(self.cards,function(a,b) return a.rank < b.rank end)
	else
		table.sort(self.cards,function(a,b) return a.rank > b.rank end)
	end
	
end

function CardGroup:SortCardsBySuit(dir)

	if (not self) then print("Error> CardGroup:SortCardsBySuit() - no group") return end
	
	if (dir and dir < 0) then
		table.sort(self.cards,function(a,b) if a.suit == b.suit then return a.rank < b.rank end return a.suit < b.suit end)
	else
		table.sort(self.cards,function(a,b) if a.suit == b.suit then return a.rank > b.rank end return a.suit > b.suit end)
	end
	
end

function CardGroup:SortCardsRandomly()

	if (not self) then print("Error> CardGroup:SortCardsRandomly() - no group") return end
	
	table.sort(self.cards,function(a,b) if math.random(2) == 1 then return a.rank < b.rank end return b.rank < a.rank end)
	
end

function CardGroup:SortCardsByFunction(func,dir)

	if (not func) then print("Error - CardGroup:SortCardsByFunction() - no func") return end

	if (dir and dir < 0) then
		table.sort(self.cards,function(a,b) return func(b,a) end)
	else
		table.sort(self.cards,func)
	end

end

function CardGroup:GetHighestRankingCard()

	hrank = 0
	hnum = 0
	for i = 1,self.num_cards do
		if (self.cards[i].rank > hrank) then hnum = i; hrank = self.cards[i].rank end
	end
	
	if (hnum == 0) then return nil end
	
	ret_card = RS(self.cards[hnum].rank,self.cards[hnum].suit)
	
	return ret_card
end

function CardGroup:GetHighestRankingCardOfSuit(s)

	hrank = 0
	hnum = 0
	for i = 1,self.num_cards do
		if (self.cards[i].rank > hrank and self.cards[i].suit == s) then hnum = i; hrank = self.cards[i].rank end
	end
	
	if (hnum == 0) then return nil end
	
	ret_card = RS(self.cards[hnum].rank,self.cards[hnum].suit)
	
	return ret_card
end

function CardGroup:GetARandomCard()

	if (self.num_cards < 1) then return nil end
	if (self.num_cards == 1) then return self.cards[1] end
	return self.cards[math.random(1,self.num_cards)]

end

function CardGroup:Display(info)

	str = ""
	if (info) then str = str .. info end
	
	if (self.name) then str = str .. self.name
	else                str = str .. "Unamed Group"
	end
	str = str .. " [" .. tostring(self.num_cards) .. "]: "
	
	for i=1,self.num_cards do
		c = self.cards[i]
		if (c) then
			if (c.rank and c.rank >= 1 and c.rank <= ace) then str = str .. rank_chars[c.rank] else str = str .. "?" end
			if (c.suit and c.suit >= 1 and c.suit <= 4)   then str = str .. suit_chars[c.suit] else str = str .. "?" end
		else
			str = str .. "--"
		end
		if (i < self.num_cards) then str = str .. " " end
	end
	
	print(str)
	
end

function CardGroup:DisplayTopCard(info)

	str = ""
	if (info) then str = str .. info end
	
	if (self.name) then str = str .. self.name
	else                str = str .. "Unamed Group"
	end
	str = str .. " [" .. tostring(self.num_cards) .. "]: "
	
	if (self.num_cards > 0) then
		c = self.cards[self.num_cards]
		-- str = str .. GetCardString(c)
		if (c) then
			if (c.rank and c.rank >= 1 and c.rank <= ace) then str = str .. rank_chars[c.rank] else str = str .. "?" end
			if (c.suit and c.suit >= 1 and c.suit <= 4)   then str = str .. suit_chars[c.suit] else str = str .. "?" end
		else
			str = str .. "--"
		end
	end
	
	print(str)
	
end


-----------------------------------------------------------------------------------------------------------
-- IsCardRS
-----------------------------------------------------------------------------------------------------------

function IsCardRS(card,rank,suit)

	if (not card) then print("IsCardRS(): no card") return end
	if (not rank or not suit) then print("IsCardRS(): no rank or suit") return end

	if (card.suit == suit and card.rank == rank) then return true end
	return false
	
end

-----------------------------------------------------------------------------------------------------------
-- IsCard
-----------------------------------------------------------------------------------------------------------

function IsCard(c1,c2)

	if not c1 or not c2 then print("IsCard(): no cards (",c1,c2,")") return end

	if c1.suit == c2.suit and c1.rank == c2.rank then return true end
	return false
	
end

-----------------------------------------------------------------------------------------------------------
-- GetCardString
-----------------------------------------------------------------------------------------------------------

function GetCardString(c)

	if (not c) then return "(no card)" end

	str = rank_strs[c.rank] .. " of " .. suit_strs[c.suit]
	return str

end

-----------------------------------------------------------------------------------------------------------
-- GetCardStringShort
-----------------------------------------------------------------------------------------------------------

function GetCardStringShort(c)

	if (not c) then return("--") end

	return rank_chars[c.rank] .. suit_chars[c.suit]

end

-----------------------------------------------------------------------------------------------------------
-- DisplayCard
-----------------------------------------------------------------------------------------------------------

function DisplayCard(c)

	print(GetCardStringShort(c))

end

-----------------------------------------------------------------------------------------------------------
-- DisplayCardWithPrefix
-----------------------------------------------------------------------------------------------------------

function DisplayCardWithPrefix(prefix,c)

	if (not c) then print("--") end

	print(prefix .. rank_chars[c.rank] .. suit_chars[c.suit])

end
